/* Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "ap_provider.h"
#include "httpd.h"
#include "http_config.h"
#include "apr_tables.h"
#include "http_log.h"
#include "apr_dbd.h"
#include "mod_dbd.h"
#include "apr_strings.h"
#include "mod_auth.h"
#include "apr_md5.h"

module AP_MODULE_DECLARE_DATA authn_bugzilla_module;

typedef struct {
    const char *user;
    const char *realm;
} authn_bugzilla_conf;

/* optional function - look it up once in post_config */
static ap_dbd_t *(*authn_bugzilla_acquire_fn)(request_rec*) = NULL;
static void (*authn_bugzilla_prepare_fn)(server_rec*, const char*, const char*) = NULL;

static void *authn_bugzilla_cr_conf(apr_pool_t *pool, char *dummy)
{
    authn_bugzilla_conf *ret = apr_pcalloc(pool, sizeof(authn_bugzilla_conf));
    return ret;
}

static void *authn_bugzilla_merge_conf(apr_pool_t *pool, void *BASE, void *ADD)
{
    authn_bugzilla_conf *add = ADD;
    authn_bugzilla_conf *base = BASE;
    authn_bugzilla_conf *ret = apr_palloc(pool, sizeof(authn_bugzilla_conf));
    ret->user = (add->user == NULL) ? base->user : add->user;
    ret->realm = (add->realm == NULL) ? base->realm : add->realm;
    return ret;
}

static const char *authn_bugzilla_prepare(cmd_parms *cmd, void *cfg, const char *query)
{
    static unsigned int label_num = 0;
    char *label;

    if (authn_bugzilla_prepare_fn == NULL) {
        authn_bugzilla_prepare_fn = APR_RETRIEVE_OPTIONAL_FN(ap_dbd_prepare);
        if (authn_bugzilla_prepare_fn == NULL) {
            return "You must load mod_dbd to enable AuthBugzilla functions";
        }
        authn_bugzilla_acquire_fn = APR_RETRIEVE_OPTIONAL_FN(ap_dbd_acquire);
    }
    label = apr_psprintf(cmd->pool, "authn_bugzilla_%d", ++label_num);

    authn_bugzilla_prepare_fn(cmd->server, query, label);

    /* save the label here for our own use */
    return ap_set_string_slot(cmd, cfg, label);
}

static authn_status authn_bugzilla_password(request_rec *r, const char *user,
                                       const char *password)
{
    apr_status_t rv;
    const char *dbd_password = NULL;
    const char *login_name = NULL;
    const char *extern_id = NULL;
    const char *realname = NULL;
    apr_dbd_prepared_t *statement;
    apr_dbd_results_t *res = NULL;
    apr_dbd_row_t *row = NULL;

    authn_bugzilla_conf *conf = ap_get_module_config(r->per_dir_config,
                                                &authn_bugzilla_module);
    ap_dbd_t *dbd = authn_bugzilla_acquire_fn(r);
    if (dbd == NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                      "Error looking up %s in database", user);
        return AUTH_GENERAL_ERROR;
    }

    if (conf->user == NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "No DBD Authn configured!");
        return AUTH_GENERAL_ERROR;
    }

    statement = apr_hash_get(dbd->prepared, conf->user, APR_HASH_KEY_STRING);
    if (statement == NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "No DBD Authn configured!");
        return AUTH_GENERAL_ERROR;
    }
    if (apr_dbd_pvselect(dbd->driver, r->pool, dbd->handle, &res, statement,
                              0, user, NULL) != 0) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                      "Error looking up %s in database", user);
        return AUTH_GENERAL_ERROR;
    }
    for (rv = apr_dbd_get_row(dbd->driver, r->pool, res, &row, -1);
         rv != -1;
         rv = apr_dbd_get_row(dbd->driver, r->pool, res, &row, -1)) {
        if (rv != 0) {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r,
                      "Error looking up %s in database", user);
            return AUTH_GENERAL_ERROR;
        }
        if (dbd_password == NULL) {
            dbd_password = apr_dbd_get_entry(dbd->driver, row, 0);
            login_name = apr_dbd_get_entry(dbd->driver, row, 1);
            extern_id = apr_dbd_get_entry(dbd->driver, row, 2);
            realname = apr_dbd_get_entry(dbd->driver, row, 3);
        }
        /* we can't break out here or row won't get cleaned up */
    }

    if (!dbd_password) {
        return AUTH_USER_NOT_FOUND;
    }

    rv = apr_password_validate(password, dbd_password);

    if (rv != APR_SUCCESS) {
        return AUTH_DENIED;
    }

    if (login_name != NULL) {
        apr_table_set(r->subprocess_env, "AUTHN_BUGZILLA_LOGIN_NAME", login_name);
    }
    if (extern_id != NULL) {
    	apr_table_set(r->subprocess_env, "AUTHN_BUGZILLA_EXTERN_ID", extern_id);
    }
    if (realname != NULL) {
    	apr_table_set(r->subprocess_env, "AUTHN_BUGZILLA_REALNAME", realname);
    }
    
    return AUTH_GRANTED;
}

static authn_status authn_bugzilla_realm(request_rec *r, const char *user,
                                    const char *realm, char **rethash)
{
    apr_status_t rv;
    const char *dbd_hash = NULL;
    apr_dbd_prepared_t *statement;
    apr_dbd_results_t *res = NULL;
    apr_dbd_row_t *row = NULL;

    authn_bugzilla_conf *conf = ap_get_module_config(r->per_dir_config,
                                                &authn_bugzilla_module);
    ap_dbd_t *dbd = authn_bugzilla_acquire_fn(r);
    if (dbd == NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                      "Error looking up %s in database", user);
        return AUTH_GENERAL_ERROR;
    }
    if (conf->realm == NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "No DBD Authn configured!");
        return AUTH_GENERAL_ERROR;
    }
    statement = apr_hash_get(dbd->prepared, conf->realm, APR_HASH_KEY_STRING);
    if (statement == NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "No DBD Authn configured!");
        return AUTH_GENERAL_ERROR;
    }
    if (apr_dbd_pvselect(dbd->driver, r->pool, dbd->handle, &res, statement,
                              0, user, realm, NULL) != 0) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                      "Error looking up %s:%s in database", user, realm);
        return AUTH_GENERAL_ERROR;
    }
    for (rv = apr_dbd_get_row(dbd->driver, r->pool, res, &row, -1);
         rv != -1;
         rv = apr_dbd_get_row(dbd->driver, r->pool, res, &row, -1)) {
        if (rv != 0) {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r,
                      "Error looking up %s in database", user);
            return AUTH_GENERAL_ERROR;
        }
        if (dbd_hash == NULL) {
            dbd_hash = apr_dbd_get_entry(dbd->driver, row, 0);
        }
        /* we can't break out here or row won't get cleaned up */
    }

    if (!dbd_hash) {
        return AUTH_USER_NOT_FOUND;
    }

    *rethash = apr_pstrdup(r->pool, dbd_hash);
    return AUTH_USER_FOUND;
}

static void authn_bugzilla_register_hooks(apr_pool_t *p)
{
    static const authn_provider authn_bugzilla_provider = {
        &authn_bugzilla_password,
        &authn_bugzilla_realm
    };

    ap_register_provider(p, AUTHN_PROVIDER_GROUP, "bugzilla", "0", &authn_bugzilla_provider);
}

static const command_rec authn_bugzilla_cmds[] =
{
    AP_INIT_TAKE1("AuthBugzillaUserPWQuery", authn_bugzilla_prepare,
                  (void *)APR_OFFSETOF(authn_bugzilla_conf, user), ACCESS_CONF,
                  "Query used to fetch password for user"),
    AP_INIT_TAKE1("AuthBugzillaUserRealmQuery", authn_bugzilla_prepare,
                  (void *)APR_OFFSETOF(authn_bugzilla_conf, realm), ACCESS_CONF,
                  "Query used to fetch password for user+realm"),
    {NULL}
};

/* Dispatch list for API hooks */
module AP_MODULE_DECLARE_DATA authn_bugzilla_module = {
    STANDARD20_MODULE_STUFF, 
    authn_bugzilla_cr_conf,                  /* create per-dir    config structures */
    authn_bugzilla_merge_conf,               /* merge  per-dir    config structures */
    NULL,                  /* create per-server config structures */
    NULL,                  /* merge  per-server config structures */
    authn_bugzilla_cmds,                     /* table of config file commands       */
    authn_bugzilla_register_hooks  /* register hooks                      */
};

